@@ -42,6 +42,19 @@
 extern gboolean GdmAllowRoot;
 extern gboolean GdmAllowRemoteRoot;
 extern gint GdmRetryDelay;
+extern gboolean GdmDisplayLastLogin;
+
+static char *selected_user = NULL;
+
+void
+gdm_verify_select_user (const char *user)
+{
+	g_free (selected_user);
+	if (ve_string_empty (user))
+		selected_user = NULL;
+	else
+		selected_user = g_strdup (user);
+}
 
 static void
 print_cant_auth_errbox (void)
@@ -90,22 +103,46 @@ gdm_verify_user (GdmDisplay *d, const char *username, const gchar *display, gboo
     gchar *login, *passwd, *ppasswd;
     struct passwd *pwent;
     struct spwd *sp;
+#if defined(HAVE_PASSWDEXPIRED) && defined(HAVE_CHPASS) \
+    || defined(HAVE_LOGINRESTRICTIONS)
+    gchar *message = NULL;
+#endif
+#if defined(HAVE_PASSWDEXPIRED) && defined(HAVE_CHPASS)
+    gchar *info_msg = NULL, *response = NULL;
+    gint reEnter, ret;
+#endif
 
     if (local)
 	    gdm_slave_greeter_ctl_no_ret (GDM_STARTTIMER, "");
 
     if (username == NULL) {
+authenticate_again:
 	    /* Ask for the user's login */
+	    gdm_verify_select_user (NULL);
 	    gdm_slave_greeter_ctl_no_ret (GDM_MSG, _("Please enter your username"));
 	    login = gdm_slave_greeter_ctl (GDM_PROMPT, _("Username:"));
 	    if (login == NULL ||
 		gdm_slave_greeter_check_interruption ()) {
-		    if (local)
-			    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
-		    g_free (login);
-		    return NULL;
+		    if ( ! ve_string_empty (selected_user)) {
+			    /* user selected */
+			    g_free (login);
+			    login = selected_user;
+			    selected_user = NULL;
+		    } else {
+			    /* some other interruption */
+			    if (local)
+				    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
+			    g_free (login);
+			    return NULL;
+		    }
 	    }
 	    gdm_slave_greeter_ctl_no_ret (GDM_MSG, "");
+
+	    if (GdmDisplayLastLogin) {
+		    char *info = gdm_get_last_info (s);
+		    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, info);
+		    g_free (info);
+	    }
     } else {
 	    login = g_strdup (username);
     }
@@ -157,7 +194,7 @@ gdm_verify_user (GdmDisplay *d, const char *username, const gchar *display, gboo
 	    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
 
     if (pwent == NULL) {
-	    sleep (GdmRetryDelay);
+	    gdm_sleep_no_signal (GdmRetryDelay);
 	    gdm_error (_("Couldn't authenticate user \"%s\""), login);
 
 	    print_cant_auth_errbox ();
@@ -171,7 +208,7 @@ gdm_verify_user (GdmDisplay *d, const char *username, const gchar *display, gboo
     /* Check whether password is valid */
     if (ppasswd == NULL || (ppasswd[0] != '\0' &&
 			    strcmp (crypt (passwd, ppasswd), ppasswd) != 0)) {
-	    sleep (GdmRetryDelay);
+	    gdm_sleep_no_signal (GdmRetryDelay);
 	    gdm_error (_("Couldn't authenticate user \"%s\""), login);
 
 	    print_cant_auth_errbox ();
@@ -198,6 +235,30 @@ gdm_verify_user (GdmDisplay *d, const char *username, const gchar *display, gboo
 	    return NULL;
     }
 
+#ifdef HAVE_LOGINRESTRICTIONS
+
+    /* Check with the 'loginrestrictions' function
+       if the user has been disallowed */
+    if (loginrestrictions (login, 0, NULL, &message) != 0) {
+	    gdm_error (_("User %s not allowed to log in"), login);
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
+					  _("\nThe system administrator "
+					    "has disabled your "
+					    "account."));
+	    g_free (login);
+	    g_free (passwd);
+	    g_free (ppasswd);
+	    if (message != NULL)
+		    free (message);
+	    return NULL;
+    }
+    
+    if (message != NULL)
+	    free (message);
+    message = NULL;
+
+#else /* ! HAVE_LOGINRESTRICTIONS */
+
     /* check for the standard method of disallowing users */
     if (pwent->pw_shell != NULL &&
 	(strcmp (pwent->pw_shell, "/sbin/nologin") == 0 ||
@@ -215,9 +276,18 @@ gdm_verify_user (GdmDisplay *d, const char *username, const gchar *display, gboo
 	    g_free (ppasswd);
 	    return NULL;
     }	
+
+#endif /* HAVE_LOGINRESTRICTIONS */
+
     g_free (passwd);
     g_free (ppasswd);
 
+    if ( ! gdm_slave_check_user_wants_to_log_in (login)) {
+	    g_free (login);
+	    login = NULL;
+	    goto authenticate_again;
+    }
+
     if ( ! gdm_setup_gids (login, pwent->pw_gid)) {
 	    gdm_error (_("Cannot set user group for %s"), login);
 	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
@@ -228,6 +298,74 @@ gdm_verify_user (GdmDisplay *d, const char *username, const gchar *display, gboo
 	    return NULL;
     }
 
+#if defined(HAVE_PASSWDEXPIRED) && defined (HAVE_CHPASS)
+
+    switch (passwdexpired (login, &info_msg)) {
+    case 1 :
+	    gdm_error (_("Password of %s has expired"), login);
+	    gdm_error_box (d, GTK_MESSAGE_ERROR,
+			   _("You are required to change your password.\n"
+			     "Please choose a new one."));
+	    g_free (info_msg);
+
+	    do {
+		    ret = chpass (login, response, &reEnter, &message);
+		    g_free (response);
+
+		    if (ret != 1) {
+			    if (ret != 0) {
+				    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
+								  _("\nCannot change your password, "
+								    "you will not be able to log in, "
+								    "please try again later or contact "
+								    "your system administrator."));
+			    } else if ((reEnter != 0) && (message)) {
+				    response = gdm_slave_greeter_ctl (GDM_NOECHO, message);
+				    if (response == NULL)
+					    response = g_strdup ("");
+			    }
+		    }
+
+		    g_free(message);
+		    message = NULL;
+
+	    } while ( ((reEnter != 0) && (ret == 0))
+		      || (ret ==1) );
+
+	    g_free (response);
+	    g_free (message);
+
+	    if ((ret != 0) || (reEnter != 0)) {
+		    return NULL;
+	    }
+
+	    break;
+
+    case 2 :
+	    gdm_error (_("Password of %s has expired"), login);
+	    gdm_error_box (d, GTK_MESSAGE_ERROR,
+			   _("Your password has expired.\n"
+			     "Only a system administrator can now change it"));
+	    g_free (info_msg);
+	    return NULL;
+	    break;    
+
+    case -1 :
+	    gdm_error (_("Internal error on passwdexpired"));
+	    gdm_error_box (d, GTK_MESSAGE_ERROR,
+			   _("An internal error occured, you will not be able to log in.\n"
+			     "Please try again later or contact your system administrator."));
+	    g_free (info_msg);
+	    return NULL;
+	    break;    
+
+    default :
+	    g_free (info_msg);
+	    break;
+    }
+
+#endif /* HAVE_PASSWDEXPIRED && HAVE_CHPASS */
+
     return login;
 }
 
